from netmiko.ubiquiti.edge_ssh import UbiquitiEdgeSSH
from netmiko.ubiquiti.edgerouter_ssh import UbiquitiEdgeRouterSSH
from netmiko.ubiquiti.unifiswitch_ssh import UbiquitiUnifiSwitchSSH

__all__ = [
    "UbiquitiEdgeRouterSSH",
    "UbiquitiEdgeSSH",
    "UbiquitiUnifiSwitchSSH",
]
